@@ -1,5 +1,3 @@
-# coding=UTF-8
-
 # Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
 # Copyright (C) 2009-2013 Kai Willadsen <kai.willadsen@gmail.com>
 #
@@ -26,6 +24,7 @@ import stat
 import sys
 from collections import namedtuple
 from decimal import Decimal
+from mmap import ACCESS_COPY, mmap
 
 from gi.repository import Gdk
 from gi.repository import Gio
@@ -39,7 +38,7 @@ from meld import tree
 from meld.conf import _
 from meld.iohelpers import trash_or_confirm
 from meld.melddoc import MeldDoc
-from meld.misc import all_same, with_focused_pane
+from meld.misc import all_same, apply_text_filters, with_focused_pane
 from meld.recent import RecentType
 from meld.settings import bind_settings, meldsettings, settings
 from meld.treehelpers import refocus_deleted_path, tree_path_as_tuple
@@ -85,11 +84,63 @@ CHUNK_SIZE = 4096
 
 
 def remove_blank_lines(text):
-    splits = text.splitlines()
-    lines = text.splitlines(True)
-    blanks = set([i for i, l in enumerate(splits) if not l])
-    lines = [l for i, l in enumerate(lines) if i not in blanks]
-    return b''.join(lines)
+    """
+    Remove blank lines from text.
+    And normalize line ending
+    """
+    return b'\n'.join(filter(bool, text.splitlines()))
+
+
+def _files_contents(files, stats):
+    mmaps = []
+    is_bin = False
+    contents = [b'' for file_obj in files]
+
+    for index, file_and_stat in enumerate(zip(files, stats)):
+        file_obj, stat_ = file_and_stat
+        # use mmap for files with size > CHUNK_SIZE
+        data = b''
+        if stat_.size > CHUNK_SIZE:
+            data = mmap(file_obj.fileno(), 0, access=ACCESS_COPY)
+            mmaps.append(data)
+        else:
+            data = file_obj.read()
+        contents[index] = data
+
+        # Rough test to see whether files are binary.
+        chunk_size = min([stat_.size, CHUNK_SIZE])
+        if b"\0" in data[:chunk_size]:
+            is_bin = True
+
+    return contents, mmaps, is_bin
+
+
+def _contents_same(contents, file_size):
+    other_files_index = list(range(1, len(contents)))
+    chunk_range = zip(
+        range(0, file_size, CHUNK_SIZE),
+        range(CHUNK_SIZE, file_size + CHUNK_SIZE, CHUNK_SIZE)
+    )
+
+    for start, end in chunk_range:
+        chunk = contents[0][start:end]
+        for index in other_files_index:
+            if not chunk == contents[index][start:end]:
+                return Different
+
+
+def _normalize(contents, ignore_blank_lines, regexes = ()):
+    contents = (bytes(c) for c in contents)
+    # For probable text files, discard newline differences to match
+    if ignore_blank_lines:
+        contents = (remove_blank_lines(c) for c in contents)
+    else:
+        contents = (b"\n".join(c.splitlines()) for c in contents)
+
+    if regexes:
+        contents = (apply_text_filters(c, regexes) for c in contents)
+
+    return contents
 
 
 def _files_same(files, regexes, comparison_args):
@@ -107,14 +158,16 @@ def _files_same(files, regexes, comparison_args):
         return Same
 
     files = tuple(files)
-    regexes = tuple(regexes)
     stats = tuple([StatItem._make(os.stat(f)) for f in files])
 
     shallow_comparison = comparison_args['shallow-comparison']
     time_resolution_ns = comparison_args['time-resolution']
     ignore_blank_lines = comparison_args['ignore_blank_lines']
+    apply_text_filters = comparison_args['apply-text-filters']
 
-    need_contents = comparison_args['apply-text-filters']
+    need_contents = ignore_blank_lines or apply_text_filters
+
+    regexes = tuple(regexes) if apply_text_filters else ()
 
     # If all entries are directories, they are considered to be the same
     if all([stat.S_ISDIR(s.mode) for s in stats]):
@@ -131,8 +184,9 @@ def _files_same(files, regexes, comparison_args):
         )
         return DodgySame if all_same_timestamp else Different
 
+    same_size = all_same([s.size for s in stats])
     # If there are no text filters, unequal sizes imply a difference
-    if not need_contents and not all_same([s.size for s in stats]):
+    if not need_contents and not same_size:
         return Different
 
     # Check the cache before doing the expensive comparison
@@ -142,38 +196,31 @@ def _files_same(files, regexes, comparison_args):
         return cache.result
 
     # Open files and compare bit-by-bit
-    contents = [[] for f in files]
     result = None
 
     try:
-        handles = [open(f, "rb") for f in files]
+        mmaps = []
+        handles = [open(file_path, "rb") for file_path in files]
         try:
-            data = [h.read(CHUNK_SIZE) for h in handles]
+            contents, mmaps, is_bin = _files_contents(handles, stats)
 
-            # Rough test to see whether files are binary. If files are guessed
-            # to be binary, we don't examine contents for speed and space.
-            if any(b"\0" in d for d in data):
-                need_contents = False
+            # compare files chunk-by-chunk
+            if same_size:
+                result = _contents_same(contents, stats[0].size)
+            else:
+                result = Different
 
-            while True:
-                if all_same(data):
-                    if not data[0]:
-                        break
-                else:
-                    result = Different
-                    if not need_contents:
-                        break
-
-                if need_contents:
-                    for i in range(len(data)):
-                        contents[i].append(data[i])
-
-                data = [h.read(CHUNK_SIZE) for h in handles]
+            # normalize and compare files again
+            if result == Different and need_contents and not is_bin:
+                contents = _normalize(contents, ignore_blank_lines, regexes)
+                result = SameFiltered if all_same(contents) else Different
 
         # Files are too large; we can't apply filters
         except (MemoryError, OverflowError):
             result = DodgySame if all_same(stats) else DodgyDifferent
         finally:
+            for m in mmaps:
+                m.close()
             for h in handles:
                 h.close()
     except IOError:
@@ -183,18 +230,6 @@ def _files_same(files, regexes, comparison_args):
     if result is None:
         result = Same
 
-    if result == Different and need_contents:
-        contents = [b"".join(c) for c in contents]
-        # For probable text files, discard newline differences to match
-        # file comparisons.
-        contents = [b"\n".join(c.splitlines()) for c in contents]
-
-        contents = [misc.apply_text_filters(c, regexes) for c in contents]
-
-        if ignore_blank_lines:
-            contents = [remove_blank_lines(c) for c in contents]
-        result = SameFiltered if all_same(contents) else Different
-
     _cache[cache_key] = CacheResult(stats, result)
     return result
 
@@ -208,11 +243,18 @@ COL_EMBLEM, COL_EMBLEM_SECONDARY, COL_SIZE, COL_TIME, COL_PERMS, COL_END = \
 
 class DirDiffTreeStore(tree.DiffTreeStore):
     def __init__(self, ntree):
-        tree.DiffTreeStore.__init__(
-            self, ntree, [str, str, object, object, object])
+        super().__init__(ntree, [str, str, int, float, int])
+
+    def add_error(self, parent, msg, pane):
+        defaults = {
+            COL_TIME: -1.0,
+            COL_SIZE: -1,
+            COL_PERMS: -1
+        }
+        super().add_error(parent, msg, pane, defaults)
 
 
-class CanonicalListing(object):
+class CanonicalListing:
     """Multi-pane lists with canonicalised matching and error detection"""
 
     def __init__(self, n, canonicalize=None):
@@ -257,7 +299,7 @@ class DirDiff(MeldDoc, Component):
         ('ignore-blank-lines', 'ignore-blank-lines'),
     )
 
-    apply_text_filters = GObject.property(
+    apply_text_filters = GObject.Property(
         type=bool,
         nick="Apply text filters",
         blurb=(
@@ -265,30 +307,30 @@ class DirDiff(MeldDoc, Component):
             "should be applied when comparing file contents"),
         default=False,
     )
-    ignore_blank_lines = GObject.property(
+    ignore_blank_lines = GObject.Property(
         type=bool,
         nick="Ignore blank lines",
         blurb="Whether to ignore blank lines when comparing file contents",
         default=False,
     )
-    ignore_symlinks = GObject.property(
+    ignore_symlinks = GObject.Property(
         type=bool,
         nick="Ignore symbolic links",
         blurb="Whether to follow symbolic links when comparing folders",
         default=False,
     )
-    shallow_comparison = GObject.property(
+    shallow_comparison = GObject.Property(
         type=bool,
         nick="Use shallow comparison",
         blurb="Whether to compare files based solely on size and mtime",
         default=False,
     )
-    status_filters = GObject.property(
+    status_filters = GObject.Property(
         type=GObject.TYPE_STRV,
         nick="File status filters",
         blurb="Files with these statuses will be shown by the comparison.",
     )
-    time_resolution = GObject.property(
+    time_resolution = GObject.Property(
         type=int,
         nick="Time resolution",
         blurb="When comparing based on mtime, the minimum difference in "
@@ -311,6 +353,7 @@ class DirDiff(MeldDoc, Component):
 
     state_actions = {
         tree.STATE_NORMAL: ("normal", "ShowSame"),
+        tree.STATE_NOCHANGE: ("normal", "ShowSame"),
         tree.STATE_NEW: ("new", "ShowNew"),
         tree.STATE_MODIFIED: ("modified", "ShowModified"),
     }
@@ -435,12 +478,16 @@ class DirDiff(MeldDoc, Component):
         self.connect("notify::ignore-blank-lines", self.update_comparator)
         self.connect("notify::apply-text-filters", self.update_comparator)
 
+        # The list copying and state_filters reset here is because the action
+        # toggled callback modifies the state while we're constructing it.
         self.state_filters = []
-        for s in self.state_actions:
-            if self.state_actions[s][0] in self.props.status_filters:
-                self.state_filters.append(s)
-                action_name = self.state_actions[s][1]
+        state_filters = []
+        status_filters = list(self.props.status_filters)
+        for state, (filter_name, action_name) in self.state_actions.items():
+            if filter_name in status_filters:
+                state_filters.append(state)
                 self.actiongroup.get_action(action_name).set_active(True)
+        self.state_filters = state_filters
 
         self._scan_in_progress = 0
 
@@ -731,11 +778,11 @@ class DirDiff(MeldDoc, Component):
 
                 for e in entries:
                     try:
-                        if not isinstance(e, str):
-                            e = e.decode('utf8')
-                    except UnicodeDecodeError:
-                        approximate_name = e.decode('utf8', 'replace')
-                        encoding_errors.append((pane, approximate_name))
+                        e.encode('utf8')
+                    except UnicodeEncodeError:
+                        invalid = e.encode('utf8', 'surrogatepass')
+                        printable = invalid.decode('utf8', 'backslashreplace')
+                        encoding_errors.append((pane, printable))
                         continue
 
                     try:
@@ -955,12 +1002,12 @@ class DirDiff(MeldDoc, Component):
                             (_("_Replace"), Gtk.ResponseType.OK),
                         ]
                         replace = misc.modal_dialog(
-                            primary=_(u"Replace folder “%s”?") % folder_name,
+                            primary=_("Replace folder “%s”?") % folder_name,
                             secondary=_(
-                                u"Another folder with the same name already "
-                                u"exists in “%s”.\n"
-                                u"If you replace the existing folder, all "
-                                u"files in it will be lost.") % parent_name,
+                                "Another folder with the same name already "
+                                "exists in “%s”.\n"
+                                "If you replace the existing folder, all "
+                                "files in it will be lost.") % parent_name,
                             buttons=dialog_buttons,
                             messagetype=Gtk.MessageType.WARNING,
                         )
@@ -1029,7 +1076,19 @@ class DirDiff(MeldDoc, Component):
 
             busy = self._scan_in_progress > 0
 
+            is_single_foldable_row = False
+            if (selection.count_selected_rows() == 1):
+                path = selection.get_selected_rows()[1][0]
+                it = self.model.get_iter(path)
+                os_path = self.model.value_path(it, pane)
+                is_single_foldable_row = self.model.is_folder(
+                    it, pane, os_path)
+
             get_action("DirCompare").set_sensitive(True)
+            get_action("DirCollapseRecursively").set_sensitive(
+                is_single_foldable_row)
+            get_action("DirExpandRecursively").set_sensitive(
+                is_single_foldable_row)
             get_action("Hide").set_sensitive(True)
             get_action("DirDelete").set_sensitive(
                 is_valid and not busy)
@@ -1074,7 +1133,7 @@ class DirDiff(MeldDoc, Component):
                 # We can skip recalculation if the new cursor is between
                 # the previous/next bounds, and we weren't on a changed row
                 state = self.model.get_state(old_cursor, 0)
-                if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
+                if state not in (tree.STATE_NORMAL, tree.STATE_NOCHANGE, tree.STATE_EMPTY):
                     skip = False
                 else:
                     if self.prev_path is None and self.next_path is None:
@@ -1186,6 +1245,35 @@ class DirDiff(MeldDoc, Component):
         for row in selected:
             self.run_diff_from_iter(self.model.get_iter(row))
 
+    def on_collapse_recursive_clicked(self, action):
+        pane = self._get_focused_pane()
+        if pane is None:
+            return
+
+        root_path = self._get_selected_paths(pane)[0]
+        filter_model = Gtk.TreeModelFilter(
+            child_model=self.model, virtual_root=root_path)
+        paths_to_collapse = []
+        filter_model.foreach(self.append_paths_to_collapse, paths_to_collapse)
+        paths_to_collapse.insert(0, root_path)
+
+        for path in reversed(paths_to_collapse):
+            self.treeview[pane].collapse_row(path)
+
+    def append_paths_to_collapse(
+            self, filter_model, filter_path, filter_iter, paths_to_collapse):
+        path = filter_model.convert_path_to_child_path(filter_path)
+        paths_to_collapse.append(path)
+
+    def on_expand_recursive_clicked(self, action):
+        pane = self._get_focused_pane()
+        if pane is None:
+            return
+
+        paths = self._get_selected_paths(pane)
+        for path in paths:
+            self.treeview[pane].expand_row(path, True)
+
     def on_button_copy_left_clicked(self, button):
         self.copy_selected(-1)
 
@@ -1212,8 +1300,8 @@ class DirDiff(MeldDoc, Component):
 
     def on_filter_state_toggled(self, button):
         active_filters = [
-            a for a in self.state_actions if
-            self.actiongroup.get_action(self.state_actions[a][1]).get_active()
+            state for state, (_, action_name) in self.state_actions.items()
+            if self.actiongroup.get_action(action_name).get_active()
         ]
 
         if set(active_filters) == set(self.state_filters):
@@ -1250,7 +1338,7 @@ class DirDiff(MeldDoc, Component):
 
     def _filter_on_state(self, roots, fileslist):
         """Get state of 'files' for filtering purposes.
-           Returns STATE_NORMAL, STATE_NEW or STATE_MODIFIED
+           Returns STATE_NORMAL, STATE_NOCHANGE, STATE_NEW or STATE_MODIFIED
 
                roots - array of root directories
                fileslist - array of filename tuples of length len(roots)
@@ -1263,9 +1351,12 @@ class DirDiff(MeldDoc, Component):
             is_present = [os.path.exists(f) for f in curfiles]
             all_present = 0 not in is_present
             if all_present:
-                if self.file_compare(curfiles, regexes) in (
-                        Same, SameFiltered, DodgySame):
+                comparison_result = self.file_compare(curfiles, regexes)
+                if comparison_result in (
+                        Same, DodgySame):
                     state = tree.STATE_NORMAL
+                elif comparison_result == SameFiltered:
+                    state = tree.STATE_NOCHANGE
                 else:
                     state = tree.STATE_MODIFIED
             else:
@@ -1354,23 +1445,29 @@ class DirDiff(MeldDoc, Component):
 
         isdir = [os.path.isdir(files[j]) for j in range(self.model.ntree)]
         for j in range(self.model.ntree):
-            column_index = functools.partial(self.model.column_index, pane=j)
             if stats[j]:
                 self.model.set_path_state(
                     it, j, state, isdir[j], display_text=name_overrides[j])
                 emblem = EMBLEM_NEW if j in newest else None
-                self.model.set_value(it, column_index(COL_EMBLEM), emblem)
                 link_emblem = EMBLEM_SYMLINK if j in symlinks else None
-                self.model.set_value(
-                    it, column_index(COL_EMBLEM_SECONDARY), link_emblem)
-                self.model.set_value(it, column_index(COL_TIME), times[j])
-                self.model.set_value(it, column_index(COL_SIZE), sizes[j])
-                self.model.set_value(it, column_index(COL_PERMS), perms[j])
+                self.model.unsafe_set(it, j, {
+                    COL_EMBLEM: emblem,
+                    COL_EMBLEM_SECONDARY: link_emblem,
+                    COL_TIME: times[j],
+                    COL_SIZE: sizes[j],
+                    COL_PERMS: perms[j]
+                })
             else:
-                # TODO: More consistent state setting here would let us avoid
-                # pyobjects for column types by avoiding None use.
                 self.model.set_path_state(
                     it, j, tree.STATE_NONEXIST, any(isdir))
+                # SET time, size and perms to -1 since None of GInt is 0
+                # TODO: change it to math.nan some day
+                # https://gitlab.gnome.org/GNOME/glib/issues/183
+                self.model.unsafe_set(it, j, {
+                    COL_TIME: -1.0,
+                    COL_SIZE: -1,
+                    COL_PERMS: -1
+                })
         return different
 
     def popup_in_pane(self, pane, event):
